import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;

/**
 * An answer - several of these are displayed per question.
 *
 * @author Michael Berry (mjrb4)
 * @author Ed Parrish
 * @version 1.2  6/06/14
 */
public class Answer extends Actor {
    private static final int WIDTH = 
    	(Game.WIDTH / 2) * Game.CELL_SIZE 
    	- Game.CELL_SIZE / 2;
    private static final int HEIGHT = Game.CELL_SIZE;
    private static final int ASCENT = 16;
    private static final Color TEXT_COLOR = Color.BLACK;
    private static final Color HOVER_COLOR = Color.MAGENTA;
    private String text;
    private boolean correct;
    private boolean enabled = true;
    private Game world;
    private Question question;
    private GreenfootImage plainImg;
    private GreenfootImage hoverImg;

    /**
     * Construct a new answer.
     *
     * @param answerText The text to display.
     * @param answerCorrect Set <code>true</code> if the answer is correct,
     *        otherwise set <code>false</code>.
     */
    public Answer(String answerText, boolean answerCorrect) {
        text = answerText;
        correct = answerCorrect;
    }

    /**
     * When added to the world, draw.
     *
     * @param w the world.
     */
    public void addedToWorld(World w) {
        world = (Game) w;
    }

    /**
     * Highlight on mouse over and respond to clicks.
     */
    public void act() {
        if (enabled && Greenfoot.mouseClicked(this)) {
            if (correct) {
                question.answerResponse(true);
            } else {
                question.answerResponse(false);
            }
        }

        if (mouseOver()) {
            setImage(hoverImg);
        } else {
            setImage(plainImg);
        }
    }

    /**
     * Detect if the mouse is over this Answer.
     *
     * @return <code>true</code> on mouse over; otherwise <code>false</code>.
     */
    private boolean mouseOver() {
        MouseInfo info = Greenfoot.getMouseInfo();
        if (info == null) return false;
        int x = info.getX();
        int y = info.getY();
        int width = Game.WIDTH;
        if (x >= width / 2 + 1 && x < width && y == getY()) {
            return true;
        }
        return false;
    }

    /**
     * Returns whether the answer is correct or not.
     *
     * @return <code>true</code> if correct; otherwise <code>false</code>.
     */
    public boolean isCorrect() {
        return correct;
    }

    /**
     * Sets whether or not this Answer will respond to mouse clicks.
     *
     * @param enable Set <code>true</code> to enable, <code>false</code> to disable.
     */
    public void setEnabled(boolean enable) {
        enabled = enable;
    }

    /**
     * Returns the answer text.
     *
     * @return The answer text.
     */
    public String getText() {
        return text;
    }

    /**
     * Set a reference to the question to which this is an answer.
     *
     * @param newQuestion The question for this answer.
     */
    public void setQuestion(final Question newQuestion) {
        question = newQuestion;
        plainImg = draw(TEXT_COLOR);
        hoverImg = draw(HOVER_COLOR);
    }

    /**
     * Draw the answer to screen.
     *
     * @param colour The color in which to draw the answer.
     */
    private GreenfootImage draw(Color fgColor) {
        final int BORDER_X = 10;
        int answerNumber = question.getIndex(this);
        String answerID = Character.toString((char) ('a' + answerNumber));
        GreenfootImage image = new GreenfootImage(WIDTH, HEIGHT);
        image.setColor(fgColor);
        image.setFont(Question.TEXT_FONT);
        image.drawString(answerID + ". " + text, BORDER_X, ASCENT);
        return image;
    }

    /**
     * Returns a string representation of the object for debugging purposes.
     *
     * @return The string representation of the object.
     */
    @Override
    public String toString() {
        return text + ":" + correct;
    }
}
