import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.io.*;
import java.util.Collections;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.JOptionPane;

/**
 * Manages the operation of the game.
 *
 * @author Ed Parrish
 * @version 1.0  5/9/2014
 */
public class GameManager extends World implements ActionListener
{
    public static final int WIDTH = Game.WIDTH * Game.CELL_SIZE;
    public static final int HEIGHT = Game.HEIGHT * Game.CELL_SIZE;
    public static final GreenfootImage BTN_UP = new GreenfootImage("rectbtn-up.jpg");
    public static final GreenfootImage BTN_DWN = new GreenfootImage("rectbtn-down.jpg");
    public static final GreenfootImage BTN_HVR = new GreenfootImage("rectbtn-hover.jpg");
    public static final Font LARGE_FONT = new Font("SansSerif", Font.BOLD, 20);
    public static final Font MED_FONT = new Font("SansSerif", Font.BOLD, 16);
    private static GreenfootSound menuMusic = new GreenfootSound("menu.mid");
    private static UserInfo myInfo;
    // Controls
    private ButtonAction[] basic = {
            new ButtonAction("Moving & turning", "move-turn.txt"),
            new ButtonAction("Random behavior", "random.txt"),
            new ButtonAction("Method madness", "methods-basic.txt"),
            new ButtonAction("Basic animation", "anim-basic.txt"),
            new ButtonAction("Basic conditions", "if-basic.txt"),
            new ButtonAction("Variables & arithmetic", "var-math.txt"),
            new ButtonAction("Boss Basic", "boss-basic.txt")
        };
    private ButtonAction[] midway = {
            new ButtonAction("Piano Visualizer", "piano-viz.txt"),
            new ButtonAction("Drawing shapes", "drawing.txt"),
            new ButtonAction("Keeping score", "score.txt"),
            new ButtonAction("Basic buttons", "button-basic.txt"),
            new ButtonAction("Simple explosions", "explosions.txt"),
            new ButtonAction("Projectile motion", "projectile.txt"),
            new ButtonAction("Boss Tweener", "notyet.txt")
        };
    private ButtonAction[] advanced = {
            new ButtonAction("Explosion animation", "notyet.txt"),
            new ButtonAction("Guns 'n bullets", "notyet.txt"),
            new ButtonAction("Smoke trails", "smoke.txt"),
            new ButtonAction("Platform jumping", "notyet.txt"),
            new ButtonAction("Tile worlds", "notyet.txt"),
            new ButtonAction("Particle text", "notyet.txt"),
            new ButtonAction("Boss Wicked", "notyet.txt")
        };
    private ButtonAction[] review = {
            new ButtonAction("if statements", "if-review.txt"),
            new ButtonAction("Loop statements", "loop-basic.txt"),
            new ButtonAction("User input", "notyet.txt"),
            new ButtonAction("Collision Detection", "notyet.txt"),
            new ButtonAction("Greenfoot algorithms", "greenfoot.txt"),
            new ButtonAction("Laser turrets", "notyet.txt")
        };
    private HashMap<String, Button> buttonMap = new HashMap<String, Button>();

    /**
     * Constructor for GameManager world.
     */
    public GameManager() {
        //this(true);
        super(WIDTH, HEIGHT, 1, false);
        checkStorage();
        showMenu();
    }

    /**
     * Coordinate animated aspects of each part of the scenario.
     */
    public void act() {
        if (myInfo == null || !menuMusic.isPlaying() 
        && OptionsManager.INSTANCE.isEnabled(OptionsManager.MUSIC_ON_OFF)) {
            menuMusic.play();
        }
    }

    /**
     * Responds to button events and controls the flow of the game.
     *
     * @param c The button or other GUIComponent that created the event.
     */
    public void actionPerformed(GUIComponent c) {
        dispatchChallenges(c, basic);
        dispatchChallenges(c, midway);
        dispatchChallenges(c, advanced);
        dispatchChallenges(c, review);
        // Utility buttons
        String text = c.getText().toLowerCase();
        if (text.contains("leader")) {
            Leaderboard leaders = Leaderboard.INSTANCE;
            leaders.showTopScores(6);
            addObject(leaders, WIDTH / 2, HEIGHT / 2);
        } else if (text.contains("rewards")) {
            RewardsStore store = RewardsStore.INSTANCE;
            addObject(store, WIDTH / 2, HEIGHT / 2);
        } else if (text.contains("options")) {
            OptionsManager opman = OptionsManager.INSTANCE;
            addObject(opman, WIDTH / 2, HEIGHT / 2);
        } else if (text.contains("instructions")) {
            addObject(new Instructions(true), getWidth() / 2, getHeight() / 2 - 1);
        }
        // Test selection button
        if (c.getText().equalsIgnoreCase("Test")) {
            gameStart("test.txt");
        }
    }

    private void dispatchChallenges(GUIComponent c, ButtonAction[] list) {
        for (int i = 0; i < list.length; i++) {
            if(c.getText().equals(list[i].getButtonName())) {
                gameStart(list[i].getActionName());
            }
        }
    }

    public void checkStorage() {
        myInfo = null;
        try {
            if (UserInfo.isStorageAvailable()) {
                myInfo = UserInfo.getMyInfo();
            }
        } catch (Exception e) {
            String msg = "Error reading your data."
                + "\nCause: " + e.getCause()
                + "\nMessage: " + e.getMessage()
                + "\nContinuing as if you were not logged in.";
            JOptionPane.showMessageDialog(null, msg);
        }

    }

    /**
     * Setup the challenge board.
     */
    public void showMenu() {
        checkStorage(); // refresh in case just logged in
        //         if (UserInfo.isStorageAvailable()) {
        //             myInfo = UserInfo.getMyInfo(); // refresh in case just logged in
        //         } else {
        //             myInfo = null;
        //         }
        List<Actor> list = getObjects(null);
        removeObjects(list);
        drawBackground();
        final int START_X = 3 * 20 + 50;
        final int DELTA_X = 6 * 20;
        final int START_Y = 1 * 20 + 10;
        final int LABEL_Y = START_Y + 2 * 20;
        final int BUTTON_Y = START_Y + 5 * 20 + 10;
        addObject(new Label("Jumbled Code Games", LARGE_FONT, Color.BLUE), WIDTH / 2, START_Y);
        displayMenuColumn("Starting\nChallenges", basic, START_X, LABEL_Y);
        displayMenuColumn("Intermediate\nChallenges", midway, START_X + DELTA_X, LABEL_Y);
        displayMenuColumn("Advanced\nChallenges", advanced, START_X + DELTA_X * 2, LABEL_Y);
        displayMenuColumn("More\nChallenges", review, START_X + DELTA_X * 3, LABEL_Y);
        // Options
        final int OPTION_X = START_X + DELTA_X * 4;
        addObject(new Label("Options &\nSettings", MED_FONT), OPTION_X, LABEL_Y);
        String[] optionText = {"Leader\nboard", "Rewards store", "Options manager", "Instructions"};
        for (int i = 0; i < optionText.length; i++) {
            Button b = new Button(BTN_UP, BTN_DWN, BTN_HVR);
            b.setFont(MED_FONT);
            b.setForeground(Color.DARK_GRAY);
            b.setText(optionText[i]);
            b.setSize(new Dimension(100, 50));
            b.addActionListener(this);
            addObject(b, OPTION_X, BUTTON_Y + i * 3 * 20);
        }
        // Extra button for testing and debugging
        //         Button play = new Button("Test", MED_FONT, Color.BLUE, Color.CYAN);
        //         play.setBackgroundPressed(Color.RED);
        //         play.setBackgroundHover(Color.YELLOW);
        //         play.setSize(new Dimension(100, 50));
        //         play.addActionListener(this);
        //         addObject(play, OPTION_X, BUTTON_Y + 15 * 20); // test button
        if (UserInfo.isStorageAvailable()) {
            adjustButtonState(); // Comment out to skip levels.
        }
    }

    public void drawBackground() {
        GreenfootImage bg = getBackground();
        bg.setColor(Color.WHITE);
        bg.fill();
        drawWallpaper(bg);
        if (myInfo == null) {
//            bg.setColor(Color.RED);
//            bg.setFont(MED_FONT);
//            bg.drawString("Please login to track progress and use other features.", 20, 18);
        }
    }

    private void displayMenuColumn(String name, ButtonAction[] list, int x, int y) {
        addObject(new Label(name, MED_FONT), x, y);
    for (int i = 0; i < list.length; i++) {
            Button b = new Button(BTN_UP, BTN_DWN, BTN_HVR);
            b.setFont(MED_FONT);
            b.setForeground(Color.DARK_GRAY);
            b.setText(list[i].getButtonName());
            b.setSize(new Dimension(100, 50));
            b.addActionListener(this);
            buttonMap.put(list[i].buttonName, b);
            // TODO: following line to disable button is temporary.
            if (list[i].actionName.equals("notyet.txt")) b.setEnabled(false);
            addObject(b, x, y + 3 * 20 + i * 3 * 20);
        }
    }

    private void adjustButtonState() {
        if (myInfo == null) return; // allow casual users to play anything
        int level = Leaderboard.INSTANCE.rewardedLevels();
        if (level < 13) {
            buttonMap.get("Boss Tweener").setEnabled(false);
            buttonMap.get("if statements").setEnabled(false);
            buttonMap.get("Loop statements").setEnabled(false);
            buttonMap.get("Greenfoot algorithms").setEnabled(false);
        }
        if (level < 7) {
            buttonMap.get("Piano Visualizer").setEnabled(false);
            buttonMap.get("Drawing shapes").setEnabled(false);
            buttonMap.get("Keeping score").setEnabled(false);
            buttonMap.get("Smoke trails").setEnabled(false);
            buttonMap.get("Simple explosions").setEnabled(false);
            buttonMap.get("Projectile motion").setEnabled(false);
            buttonMap.get("Basic buttons").setEnabled(false);
        }
        if (level < 6) {
            buttonMap.get("Boss Basic").setEnabled(false);
        }
        if (level == 0) {
            buttonMap.get("Random behavior").setEnabled(false);
            buttonMap.get("Method madness").setEnabled(false);
            buttonMap.get("Basic animation").setEnabled(false);
            buttonMap.get("Basic conditions").setEnabled(false);
            buttonMap.get("Variables & arithmetic").setEnabled(false);
        }
    }

    private void gameStart(String fileName) {
        menuMusic.stop();
        Greenfoot.setWorld(new Game(fileName));
    }

    public static void setMusicEnabled(boolean on) {
        if (on) {
            menuMusic.play();
        } else {
            menuMusic.stop();
        }
    }

    public static UserInfo getUserInfo() {
        if (myInfo == null && UserInfo.isStorageAvailable()) {
            myInfo = UserInfo.getMyInfo(); // refresh in case just logged in
        }
        return myInfo;
    }

    public static void drawWallpaper(GreenfootImage img) {
        OptionsManager opman = OptionsManager.INSTANCE;
        List<OptionsManager.Setting> wallpaper = opman.getWallpaper();
        for (OptionsManager.Setting setting: wallpaper) {
            String name = setting.getName().toLowerCase();
            if (name.contains("avatar")) {
                drawRandomAvatars(img, 50);
            }
            if (name.contains("bubble")) {
                drawRandomBubbles(img, 50);
            }
            if (name.contains("wombat")) {
                drawRandomWombats(img, 50);
            }
        }
    }

    private static void drawRandomAvatars(GreenfootImage img, int numImages) {
        UserInfo myInfo = getUserInfo();
        GreenfootImage avatar = null;
        if (myInfo != null) {
            avatar = myInfo.getUserImage();
        } else {
            avatar = new GreenfootImage("default.gif");
        }
        avatar.setTransparency(20);
        drawRandomImages(avatar, img, numImages);
    }

    // http://blogs.kent.ac.uk/mik/2012/07/joc-32/
    private static void drawRandomBubbles(GreenfootImage img, int numBubbles) {
        int width = img.getWidth();
        int height = img.getHeight();
        for (int i = 0; i < numBubbles; i++) {
            int x = Greenfoot.getRandomNumber(width);
            int y = Greenfoot.getRandomNumber(height);
            int r = Greenfoot.getRandomNumber(256);
            int g = Greenfoot.getRandomNumber(256);
            int b = Greenfoot.getRandomNumber(256);
            Color c = new Color(r, g, b, 20);
            int size = Greenfoot.getRandomNumber(50) + 50;
            img.setColor(c);
            img.fillOval(x, y, size, size);
        }
    }

    private static void drawRandomWombats(GreenfootImage img, int numImages) {
        GreenfootImage wombat = new GreenfootImage("wombat.png");
        wombat.setTransparency(20);
        drawRandomImages(wombat, img, numImages);
    }

    private static void drawRandomImages(GreenfootImage img, GreenfootImage bg, int numImgs) {
        int width = bg.getWidth();
        int height = bg.getHeight();
        for (int i = 0; i < numImgs; i++) {
            int x = Greenfoot.getRandomNumber(width);
            int y = Greenfoot.getRandomNumber(height);
            bg.drawImage(img, x, y);
        }
    }

    /**
     * Utility class to associate button labels and challenge files.
     *
     * @author Ed Parrish
     * @version 1.0  7/5/2014
     */
    private class ButtonAction { // implements Comparable<Pair<FIRST, SECOND>> {

        public final String buttonName;
        public final String actionName;

        public ButtonAction(String buttonName, String actionName) {
            this.buttonName = buttonName;
            this.actionName = actionName;
        }

        public String getButtonName() {
            return buttonName;
        }

        public String getActionName() {
            return actionName;
        }
    }
}
