import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Font;
import java.util.Collections;
import java.util.List;
import java.util.ArrayList;
import java.awt.Color;
import java.awt.Graphics;

/**
 * A question the user is asked.
 *
 * @author Michael Berry (mjrb4)
 * @author Ed Parrish
 * @version 1.2  6/06/14
 */
public class Question extends Actor {
    private static final int WIDTH = 
        Game.WIDTH / 2 * Game.CELL_SIZE - Game.CELL_SIZE / 2;
    private static final int HEIGHT = 
        (Game.CODE_HEIGHT + 4) * Game.CELL_SIZE - Game.CELL_SIZE / 4;
    public static final Font HEADER_FONT = Game.MED_FONT;
    public static final Font TEXT_FONT = new Font("SansSerif", Font.PLAIN, 14);
    public static final Font RESULT_FONT = new Font("SansSerif", Font.BOLD, 14);
    private static final Color HEADER_COLOR = Color.BLUE;
    private static final Color TEXT_COLOR = Color.BLACK;
    private static final Color CORRECT_COLOR = new Color(34, 139, 34); // Forest green
    private static final Color WRONG_COLOR = Color.RED;
    private static final Color BG_COLOR = Color.WHITE;
    private static final int BORDER_X = 10;
    private static final int BORDER_Y = 26;
    //private static final int LINE_LENGTH = 44;
    private static final int LINE_LENGTH = WIDTH - BORDER_X;// * 3 / 2; ;// - Game.CELL_SIZE
    private static final int LINE_SPACING = Game.CELL_SIZE;
    private String[] text;
    private List<Answer> answers;
    private String explanation = "";
    private int value;
    private static GreenfootImage img = new GreenfootImage(WIDTH, HEIGHT);
    private Game world;

    /**
     * Create a new default question.
     */
    public Question() {
        //text = GUIComponent.wordWrap("", LINE_LENGTH);
        text = new String[1];
        text[0] = "";
        answers = new ArrayList<Answer>();
        value = 0;
    }

    /**
     * Create a copy of a Question from another Question.
     *
     * @param original The Question to copy from.
     */
    public Question(final Question original) {
        text = new String[original.text.length];
        for (int i = 0; i < original.text.length; i++) {
            text[i] = original.text[i];
        }
        answers = new ArrayList(original.answers);
        explanation = original.explanation;
        value = original.value;
    }

    /**
     * Save the world reference and draw this Question.
     *
     * @param w the world.
     */
    public void addedToWorld(World w) {
        world = (Game) w;
        draw();
    }

    /**
     * Set the text for this question.
     *
     * @param newText The new text.
     */
    public void setText(String newText) {
        //text = GUIComponent.wordWrap(newText, LINE_LENGTH);
        Graphics g = getImage().getAwtImage().getGraphics();
        g.setFont(TEXT_FONT);
        text = GUIComponent.wordWrap(newText, LINE_LENGTH, g);
    }

    /**
     * Returns the text for this Question.
     *
     * @return The Question text as a String.
     */
    public String getText() {
        String str = "";
        for (String line : text) {
            str += line;
        }
        return str;
    }

    /**
     * Tests if this Question has text assigned. Often used to determine if there is
     * a complete question.
     *
     * @return <code>true</code> if this Question has text;
     *         otherwise <code>false</code>.
     */
    public boolean hasText() {
        return text != null && text.length > 0 && !text[0].equals("");
    }

    /**
     * Adds an answer for this question.
     *
     * @param answerText The answer text.
     * @param correct Set <code>true</code> if the answer is correct;
     *        otherwise <code>false</code>.
     */
    public void addAnswer(String answerText, boolean correct) {
        answers.add(new Answer(answerText, correct));
    }

    /**
     * Randomly permutes the list of answers.
     */
    public void shuffleAnswers() {
        Collections.shuffle(answers);
    }

    /**
     * Returns the index of the first occurrence of the specified answer,
     * or -1 if the answer does not exist.
     *
     * @param answer The Answer to find the index of.
     *
     * @return The index of the answer in the list of answers, or -1 if the
     * answer does not exist.
     */
    public int getIndex(Answer answer) {
        return answers.indexOf(answer);
    }

    /**
     * Set the optional explanation for this question.
     *
     * @param explanatoryText The explanation for this question.
     */
    public void setExplanation(String explanatoryText) {
        if (explanatoryText == null) explanatoryText = "";
        explanation = explanatoryText;
    }

    /**
     * Get the explanation for this question.
     *
     * @return The explanation for the question.
     */
    public String getExplanation() {
        return explanation;
    }

    /**
     * Determine whether this question has an explanation.
     *
     * @return <code>true</code> if the explanation has an answer;
     *         otherwise <code>false</code>.
     */
    public boolean hasExplanation() {
        return explanation != null;
    }

    /**
     * Sets the "value" for this question.
     *
     * @param newValue The new value for this question.
     */
    public void setValue(int newValue) {
        value = newValue;
    }

    /**
     * Returns the current value for this question.
     *
     * @return The current value for this question.
     */
    public int getValue() {
        return value;
    }

    /**
     * Responds to an answer to this question.
     *
     * @param correct Set <code>true</code> if the answer was right; other set
     * <code>false</code>.
     */
    public void answerResponse(boolean correct) {
        if (correct) {
            world.addToScore(value);
            Greenfoot.playSound("ansright.wav");
        } else {
            world.addToScore(-value);
            Greenfoot.playSound("answrong.wav");
        }
        for (Answer a: answers) {
            a.setEnabled(false);
        }
        drawExplanation(correct);
    }

    /**
     * Draw the question panel with its answers.
     */
    private void draw() {
        img.setColor(BG_COLOR);
        img.fill();
        img.setFont(HEADER_FONT);
        img.setColor(HEADER_COLOR);
        img.drawString("Bonus question (" + value + "):", BORDER_X, BORDER_Y);

        img.setFont(TEXT_FONT);
        img.setColor(TEXT_COLOR);
        for (int i = 0; i < text.length; i++) {
            img.drawString(text[i], BORDER_X, BORDER_Y + ((i + 1) * LINE_SPACING));
        }
        setImage(img);

        if (answers.size() != 0) {
            for (int i = 0; i < answers.size(); i++) {
                Answer a = answers.get(i);
                a.setQuestion(this);
                world.addObject(a, world.RIGHT_X, world.CODE_Y + text.length + i);
            }
        }
    }

    /**
     * Draw the explanation if available.
     */
    private void drawExplanation(boolean correct) {
        int startY = BORDER_Y + (text.length + answers.size() + 1) * LINE_SPACING;
        String msg = null;
        if (correct) {
            msg = "Correct!";
            img.setColor(CORRECT_COLOR);
        } else {
            msg = "Sorry, not correct.";
            img.setColor(WRONG_COLOR);
        }
        img.setFont(RESULT_FONT);
        img.drawString(msg, BORDER_X, startY);
        if (!hasExplanation()) return;
        img.setFont(TEXT_FONT);
        img.setColor(TEXT_COLOR);
        Graphics g = getImage().getAwtImage().getGraphics();
        g.setFont(TEXT_FONT);
        String[] explain = GUIComponent.wordWrap(explanation, LINE_LENGTH, g);
        for (int i = 0; i < explain.length; i++) {
            img.drawString(explain[i], BORDER_X, startY + ((i + 1) * LINE_SPACING));
        }
    }

    /**
     * Returns a string representation of the object for debugging purposes.
     *
     * @return The string representation of the object.
     */
    @Override
    public String toString() {
        String str = "Question:" + getText() + "(" + value + ")\n";
        str += "Answers:\n";
        for (Answer ans : answers) {
            str += ans + "\n";
        }
        if (explanation != null) str += explanation;
        return str;
    }
}
