import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;

/**
 * Presents a screen of text with a title and two buttons: Main Menu and Continue.
 * 
 * @author Ed Parrish 
 * @version 1.0 7/15/2014
 */
public class TextPanel extends Actor
{
    public static final GreenfootImage BTN_UP = new GreenfootImage("roundedbtn-up.png");
    public static final GreenfootImage BTN_DWN = new GreenfootImage("roundedbtn-down.png");
    public static final GreenfootImage BTN_HVR = new GreenfootImage("roundedbtn-hover.png");

    private static final int WIDTH = Game.WIDTH * Game.CELL_SIZE - 20;
    private static final int HEIGHT = Game.HEIGHT * Game.CELL_SIZE - 20;
    private static final Color BG_COLOR = Color.WHITE;
    private static final Color OUTLINE_COLOR = Color.BLUE;
    private static final Color TITLE_COLOR = Color.BLUE;
    private static final Color TEXT_COLOR = Color.BLACK;
    private Font titleFont = new Font("SansSerif", Font.BOLD, 24);
    private Font textFont = new Font("SansSerif", Font.PLAIN, 18);
    private Font buttonFont = new Font("SansSerif", Font.PLAIN, 18);
    private Game world;
    private String code;
    private String title;
    private String text;
    private Button next;
    private Button menu;

    /**
     * Default constructor used for testing.
     */
    public TextPanel() {
        this("ABC", "Test Title", "Test text\nline 2\nanother line", null);
    }

    /**
     * Constructs a text panel.
     *
     * @param titleText The title text.
     * @param mainText The body of the text to display.
     * @param obj The object responding to button clicks.
     */
    public TextPanel(String code, String titleText, String mainText, ActionListener obj) {
        this.code = code;
        title = titleText;
        text = mainText;
        next = new Button(BTN_UP, BTN_DWN, BTN_HVR);
        next.setForeground(Color.BLUE);
        next.setFont(buttonFont);
        next.setText("Continue");
        next.addActionListener(obj);
        menu = new Button(BTN_UP, BTN_DWN, BTN_HVR);
        menu.setForeground(Color.BLUE);
        menu.setFont(buttonFont);
        menu.setText("Main Menu");
        menu.addActionListener(obj);
    }

    /**
     * Draws the text and places the buttons.
     *
     * @param world The world to which the object was added.
     */
    protected void addedToWorld(World world) {
        this.world = (Game) world;
        draw();
        placeButtons();
    }

    private void draw() {
        GreenfootImage img = new GreenfootImage(WIDTH, HEIGHT);
        makeBackground(img);
        drawTitle(img);
        drawText(img);
        setImage(img);
    }
    
    private void makeBackground(GreenfootImage img) {
        img.setColor(OUTLINE_COLOR);
        img.fill();
        img.setColor(BG_COLOR);
        img.fillRect(2, 2, WIDTH - 4, HEIGHT - 4);
        GameManager.drawWallpaper(img);
    }
    
    private void drawTitle(GreenfootImage img) {
        img.setColor(TITLE_COLOR);
        img.setFont(titleFont);
        Graphics g = img.getAwtImage().getGraphics();
        FontMetrics metrics = g.getFontMetrics(titleFont);
        int adv = metrics.stringWidth(title);
        int  y = 30;
        img.drawString(title, (WIDTH - adv) / 2, y);
        UserInfo myInfo = GameManager.getUserInfo();
        int score = Leaderboard.INSTANCE.getScore(code);
        String lastInfo = null;
        if (score >= 0) {
            lastInfo = "Personal best score: " + score + " / " + world.maxScore();
        } else {
            lastInfo = "First attempt";
        }
        img.setFont(textFont);
        metrics = g.getFontMetrics(textFont);
        adv = metrics.stringWidth(lastInfo);
        y += metrics.getHeight();
        img.drawString(lastInfo, (WIDTH - adv) / 2, y);
    }
    
    private void drawText(GreenfootImage img) {
        final int LEFT_MARGIN = 20;
        final int RIGHT_MARGIN = 16;
        final int HORIZ_MARGINS = LEFT_MARGIN + RIGHT_MARGIN;
        img.setColor(TEXT_COLOR);
        img.setFont(textFont);
        Graphics g = img.getAwtImage().getGraphics();
        g.setFont(textFont); // needed for wordWrap()
        String[] lines = GUIComponent.wordWrap(text, WIDTH - HORIZ_MARGINS, g);
        int y = 80;
        FontMetrics metrics = g.getFontMetrics(textFont);
        int lineHeight = metrics.getHeight();
        for (String line: lines) {
            img.drawString(line, LEFT_MARGIN, y);
            y += lineHeight;
        }
    }
    
    private void placeButtons() {
        World w = getWorld();
        int x = Game.WIDTH - 5;
        int y = Game.HEIGHT - 3;
        w.addObject(next, x, y);
        x -= next.getWidth() / 20 + 1;
        w.addObject(menu, x, y);
    }
}
